#include <stdio.h>
#include <stdlib.h>
#include <sys/acl.h>

#include <et/com_err.h>

#include <linux/fs.h>
#include <linux/ext2_fs.h>

#include <ext2fs/ext2fs.h>

#define CHUNK 1024

struct ino_acl {
	unsigned long	ino;
	unsigned long	dir_acl;
	unsigned long	file_acl;
};

struct acl_ref {
	unsigned long	no;
	unsigned long	ref_count;
	unsigned long	acle_no;
};

static struct ino_acl *ino_acl_array = NULL;
static struct acl_ref *acl_ref_array = NULL;
static int ino_acl_size = 0;
static int ino_acl_count = 0;
static int acl_ref_size = 0;
static int acl_ref_count = 0;

extern const char *__progname;

static
void record_inode (ino_t ino, struct ext2_inode inode)
{
	struct ino_acl *p;

	if (!ino_acl_array) {
		ino_acl_size = CHUNK;
		ino_acl_array = malloc (ino_acl_size * sizeof (struct ino_acl));
		if (!ino_acl_array) {
			perror ("malloc");
			exit (32);
		}
	}
	if (ino_acl_count >= ino_acl_size) {
		ino_acl_size += CHUNK;
		p = realloc (ino_acl_array,
			     ino_acl_size * sizeof (struct ino_acl));
		if (!p) {
			perror ("realloc");
			free (ino_acl_array);
			exit (32);
		}
		ino_acl_array = p;
	}
	ino_acl_array[ino_acl_count].ino = ino;
	ino_acl_array[ino_acl_count].dir_acl = inode.i_dir_acl;
	ino_acl_array[ino_acl_count].file_acl = inode.i_file_acl;
	ino_acl_count++;
}

static
void record_acl (struct ext2_acl_header *aclh)
{
	struct acl_ref *p;

	if (!acl_ref_array) {
		acl_ref_size = CHUNK;
		acl_ref_array = malloc (acl_ref_size * sizeof (struct acl_ref));
		if (!acl_ref_array) {
			perror ("malloc");
			exit (32);
		}
	}
	if (acl_ref_count >= acl_ref_size) {
		acl_ref_size += CHUNK;
		p = realloc (acl_ref_array, acl_ref_size * sizeof (struct acl_ref));
		if (!p) {
			perror ("realloc");
			free (acl_ref_array);
			exit (32);
		}
		acl_ref_array = p;
	}
	acl_ref_array[acl_ref_count].no = aclh->aclh_no;
	acl_ref_array[acl_ref_count].ref_count = aclh->aclh_ref_count;
	acl_ref_array[acl_ref_count].acle_no = aclh->aclh_first_acle;
	acl_ref_count++;
}

static
int handle_acl_block (ext2_filsys fs, blk_t *blocknr, int blockcnt,
		      void *private)
{
	struct ext2_acl_desc *acld;
	struct ext2_acl_header *aclh;
	void *bitmap;
	char *buf = private;
	errcode_t retval;
	int i;

	retval = io_channel_read_blk (fs->io, *blocknr, 1, buf);
	if (retval)
		return BLOCK_ABORT;
	acld = (struct ext2_acl_desc *)buf;
	bitmap = (void *)(buf + sizeof (struct ext2_acl_desc));
	for (i = 2; i < EXT2_ACLH_PER_BLOCK(fs->super); i++) {
		aclh = (struct ext2_acl_header *)(buf + i * sizeof (struct ext2_acl_desc));
		if (aclh->aclh_valid == LINUX_ACL_VALID)
			record_acl (aclh);
	}
	return 0;
}

static
void list_acl (void)
{
	int i;
	int j;
	unsigned long n;

	for (i = 0; i < acl_ref_count; i++) {
		printf ("ACL %ld, refcount %ld, used by inodes:\n",
			acl_ref_array[i].no, acl_ref_array[i].ref_count);
		n = 0;
		for (j = 0; j < ino_acl_count; j++)
			if (ino_acl_array[j].dir_acl == acl_ref_array[i].no
			    || ino_acl_array[j].dir_acl == acl_ref_array[i].no) {
				printf ("\t%ld", ino_acl_array[j].ino);
				if (ino_acl_array[j].dir_acl ==
				    acl_ref_array[i].no) {
					printf (" (dir)");
					n++;
				}
				if (ino_acl_array[j].file_acl ==
				    acl_ref_array[i].no) {
					printf (" (file)");
					n++;
				}
				printf ("\n");
			}
		if (acl_ref_array[i].ref_count != n)
			printf ("!!!!! BAD REFCOUNT (%ld != %ld)\n",
				acl_ref_array[i].ref_count, n);
	}
}

void main (int argc, char *argv[])
{
	errcode_t retval;
	ext2_filsys fs = 0;
	ext2_inode_scan scan = 0;
	struct ext2_inode inode;
	ino_t ino;
	int ino_count = 0;
	char *buf;

	if (argc != 2) {
		fprintf (stderr, "Usage: %s device\n", __progname);
		exit (1);
	}
	retval = ext2fs_open(argv[1], 0, 0, 0, unix_io_manager, &fs);
	if (retval) {
		com_err (__progname, retval, "while trying to open %s",
			 argv[1]);
		exit (2);
	}
	retval = ext2fs_open_inode_scan (fs, 0, &scan);
	if (retval) {
		com_err (__progname, retval, "while opening inode scan");
		ext2fs_close (fs);
		exit (4);
	}
	retval = ext2fs_get_next_inode (scan, &ino, &inode);
	if (retval) {
		com_err (__progname, retval, "while starting inode scan");
		ext2fs_close_inode_scan (scan);
		ext2fs_close (fs);
		exit (8);
	}
	while (ino) {
		ino_count++;
		if (inode.i_dir_acl || inode.i_file_acl)
			record_inode (ino, inode);
		retval = ext2fs_get_next_inode (scan, &ino, &inode);
		if (retval) {
			com_err (__progname, retval, "while doing inode scan");
			ext2fs_close_inode_scan (scan);
			ext2fs_close (fs);
			free (ino_acl_array);
			exit (16);
		}
	}
	printf ("%d inodes scanned, %d inodes use ACL\n",
		ino_count, ino_acl_count);
	ext2fs_close_inode_scan (scan);
	buf = malloc (fs->blocksize);
	if (!buf) {
		perror ("malloc");
		ext2fs_close (fs);
		exit (32);
	}
	ext2fs_block_iterate (fs, EXT2_ACL_IDX_INO, 0, NULL, handle_acl_block,
			      (void *)buf);
	printf ("%d ACL found\n", acl_ref_count);
	ext2fs_close (fs);
	list_acl ();
	if (ino_acl_array)
		free (ino_acl_array);
	if (acl_ref_array)
		free (acl_ref_array);
	exit (0);
}
